﻿using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using MVR.FileManagementSecure;
using SimpleJSON;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

/* --- © 2022 JaxZoa <JaxZoa@gmail.com> --- */

namespace JaxPlugins
{
    public class CLS : MVRScript
    {
        public static string EXT_FULLBODY = "clp";
        public static string EXT_FULLTRIGGER = "clf";
        public static string EXT_DEST_TRIGGER = "cld";
        public static string MAIN_DIR;
        public static string FBP_DIR;
        public static string TP_DIR;

        /* Meta Target Parameters */

        protected JSONStorableFloat js_PresetLoadTransitionDuration;

        protected Atom positionTargetAtom;
        protected JSONStorableStringChooser js_PositionTargetAtom;

        protected Atom rotationTargetAtom;
        protected JSONStorableStringChooser js_RotationTargetAtom;

        protected Atom groundAtom;
        protected JSONStorableStringChooser js_GroundAtom;

        protected JSONStorableBool js_CanMove;
        protected JSONStorableStringChooser js_RotType;

        private Transform targetTransform;

        protected JSONStorableVector js_TargetPos;
        protected JSONStorableVector js_TargetRot;

        protected bool destinationReached = false;
        protected bool isIdle = true;
        public StateChanged idleStartDelegate;
        public Event destinationReachedDelegate;

        protected bool movementLock = false;

        protected JSONStorableFloat js_MaxSpeed;
        protected JSONStorableFloat js_Acceleration;
        protected JSONStorableFloat js_RotationSpeed;
        protected JSONStorableFloat js_StopDistance;
        protected JSONStorableFloat js_RotationOffset;
        protected JSONStorableBool js_CopyAllTargetRots;

        protected Vector3 currentSpeed = Vector3.zero;
        protected List<Collider> groundColliders;

        protected float movementDirection = 1;

        /* Actions parameters */

        protected List<string> triggerList = new List<string> { "Select Trigger", "OnIdleStart", "OnStepIdleStart", "OnMovementStart", "OnStanding", "OnKneeling", "OnBodyPresetLoadStart", "OnBodyPresetLoaded", "OnDestinationReached" };

        protected JSONStorableStringChooser js_TriggerEditorChooser;
        public EventTrigger trigger_OnIdleStart;
        public EventTrigger trigger_OnStepIdleStart;
        public EventTrigger trigger_OnMovementStart;
        public EventTrigger trigger_OnStanding;
        public EventTrigger trigger_OnKneeling;
        public EventTrigger trigger_OnPresetLoadStart;
        public EventTrigger trigger_OnPresetLoaded;
        public EventTrigger trigger_OnDestinationReached;

        protected List<EventTrigger> eventTriggers;

        private float destinationReachedTime = 0;
        private bool destinationReachedTriggered = true;

        public delegate void StateChanged(bool _newState);
        public delegate void Event();


        /* Body Controller parameters */

        private BodyController bodyController;

        public override void Init()
        {
            /* Meta Target Init */

            MAIN_DIR = GetPluginPath();
            FBP_DIR = MAIN_DIR + "\\FullBodyPresets";
            TP_DIR = MAIN_DIR + "\\TriggerPresets";

            SimpleTriggerHandler.LoadAssets();
            InitTriggers();

            activeSliders = new List<UIDynamicSlider>();
            activeToggles = new List<UIDynamicToggle>();
            activeButtons = new List<UIDynamicButton>();
            activePopups = new List<UIDynamicPopup>();
            activeTextFields = new List<UIDynamicTextField>();
            activeSpacers = new List<UIDynamic>();

            js_TargetPos = new JSONStorableVector(this, "TargetPosition", Vector3.zero);
            js_TargetRot = new JSONStorableVector(this, "TargetRotation", Vector3.zero);
      
            js_PositionTargetAtom = new JSONStorableStringChooser("Position Target Atom", GetAtomNames(), "None", "Pos Target Atom", Callback_TargetPositionAtomChanged);
            js_PositionTargetAtom.popupOpenCallback = Callback_TargetAtomOpened;
            RegisterStringChooser(js_PositionTargetAtom);

            js_RotationTargetAtom = new JSONStorableStringChooser("Rotation Target Atom", GetAtomNames(), "None", "Rot Target Atom", Callback_TargetRotationAtomChanged);
            js_RotationTargetAtom.popupOpenCallback = Callback_TargetAtomOpened;
            RegisterStringChooser(js_RotationTargetAtom);

            js_GroundAtom = new JSONStorableStringChooser("Ground Atom", GetAtomNames(), "None", "Ground Atom", Callback_GroundAtomChanged);
            js_GroundAtom.popupOpenCallback = Callback_GroundAtomOpened;
            RegisterStringChooser(js_GroundAtom);

            js_TriggerEditorChooser = new JSONStorableStringChooser("TriggerEditorChooser", triggerList, triggerList[0], "Trigger Editor", Callback_TriggerEditorChooser);
            RegisterStringChooser(js_TriggerEditorChooser);

            js_CanMove = new JSONStorableBool("Can Move", true);
            RegisterBool(js_CanMove);
            
            js_MaxSpeed = new JSONStorableFloat("Max Speed", 0.6f, 0.0f, 1.5f, true);
            RegisterFloat(js_MaxSpeed);
            
            js_Acceleration = new JSONStorableFloat("Acceleration", 1.5f, 0.2f, 5f, true);
            RegisterFloat(js_Acceleration);

            js_StopDistance = new JSONStorableFloat("Stop Distance", 0.25f, 0.01f, 1f, false);
            RegisterFloat(js_StopDistance);

            js_RotType = new JSONStorableStringChooser("RotationType", new List<string> { "No Rot", "Face Target", "Copy Target", "Copy Target Idle" }, "Face Target", "Rotation Type");
            RegisterStringChooser(js_RotType);            

            js_RotationSpeed = new JSONStorableFloat("Rotation Speed", 1f, 0.5f, 3f, true);
            RegisterFloat(js_RotationSpeed);

            js_RotationOffset = new JSONStorableFloat("Rotation Offset", 0, -180, 180);
            RegisterFloat(js_RotationOffset);

            js_CopyAllTargetRots = new JSONStorableBool("Copy All Rotations", false);
            RegisterBool(js_CopyAllTargetRots);

            groundColliders = new List<Collider>();

            /* Presets */

            js_PresetLoadTransitionDuration = new JSONStorableFloat("Preset Load Duration", 1, 0.2f, 5f);
            RegisterFloat(js_PresetLoadTransitionDuration);

            /* Body Controller init */

            bodyController = new BodyController_Standing(this);

            SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;

            CreateGlobalUI();
        }

        public void InitTriggers()
        {
            eventTriggers = new List<EventTrigger>();

            trigger_OnIdleStart = new EventTrigger(this, "OnIdleStart");
            trigger_OnStepIdleStart = new EventTrigger(this, "OnStepIdleStart");
            trigger_OnMovementStart = new EventTrigger(this, "OnMovementStart");
            trigger_OnStanding = new EventTrigger(this, "OnStandingStart");
            trigger_OnKneeling = new EventTrigger(this, "OnKneelingStart");
            trigger_OnPresetLoadStart = new EventTrigger(this, "OnPresetLoadStart");
            trigger_OnPresetLoaded = new EventTrigger(this, "OnPresetLoaded");
            trigger_OnDestinationReached = new EventTrigger(this, "OnDestinationReached");

            eventTriggers.Add(trigger_OnIdleStart);
            eventTriggers.Add(trigger_OnStepIdleStart);
            eventTriggers.Add(trigger_OnMovementStart);
            eventTriggers.Add(trigger_OnStanding);
            eventTriggers.Add(trigger_OnKneeling);
            eventTriggers.Add(trigger_OnPresetLoaded);
            eventTriggers.Add(trigger_OnDestinationReached);

            idleStartDelegate += Event_StartIdle;
            destinationReachedDelegate += Event_DestinationReached; 
        }

        void Start()
        {
            /* Meta Target Start*/

            GameObject _targetDummy = new GameObject("TargetDummy");

            targetTransform = _targetDummy.transform;

            FreeControllerV3 hipController = GetControllerByName("hipControl");

            if (hipController == null)
                SuperController.LogError("You need to put this script on a Person atom !");

            if (js_TargetPos.val == Vector3.zero && js_TargetRot.val == Vector3.zero)
            {
                js_TargetPos.val = hipController.transform.position;
                js_TargetPos.y = 0;

                js_TargetRot.y = hipController.transform.rotation.eulerAngles.y;
            }

            targetTransform.position = js_TargetPos.val;
            targetTransform.rotation = Quaternion.Euler(js_TargetRot.val);

            bodyController.Start();
        }

        void Update()
        {
            try
            {
                bool _idleState = isIdle;

                ComputeMetaTargetPosition();
                ComputeMetaTargetRotation();

                if (_idleState != isIdle)
                    idleStartDelegate.Invoke(isIdle);

                bodyController.Update();

                if (destinationReached && (Time.time - destinationReachedTime) > 0.75f && !destinationReachedTriggered)
                {
                    destinationReachedDelegate.Invoke();

                    destinationReachedTriggered = true;
                }
            } 
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
}

        void OnDestroy()
        {
            GameObject.Destroy(targetTransform.gameObject);

            bodyController.OnDestroy();

            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

            foreach (EventTrigger et in eventTriggers)
            {
                et.Remove();
            }
        }

        void ComputeMetaTargetPosition()
        {
            if (positionTargetAtom == null)
            {
                isIdle = true;
                currentSpeed = Vector3.zero;

                return;
            }

            Vector3 _targetAtomPos = positionTargetAtom.mainController.transform.position;
            _targetAtomPos.y = 0;

            Vector3 _deltaTarget = _targetAtomPos - targetTransform.position;

            if (_deltaTarget.magnitude > js_StopDistance.val)
            {
                currentSpeed += _deltaTarget.normalized * js_Acceleration.val * Time.deltaTime;
                currentSpeed = Vector3.ClampMagnitude(currentSpeed, js_MaxSpeed.val);

                isIdle = false;

                destinationReached = false;
                destinationReachedTriggered = false;
            }
            else
            {
                currentSpeed = Vector3.zero;

                isIdle = true;

                if (!destinationReached)
                {
                    destinationReached = true;

                    destinationReachedTime = Time.time;
                }
            }

            if (!js_CanMove.val || movementLock)
            {
                currentSpeed = Vector3.zero;
                isIdle = true;
            }

            targetTransform.position = targetTransform.position + (currentSpeed * Time.deltaTime);

            js_TargetPos.val = targetTransform.position;
        }

        void ComputeMetaTargetRotation()
        {
            movementDirection = 1;

            if (js_RotType.val == "No Rot")
                return;

            if (positionTargetAtom == null && rotationTargetAtom == null)
                return;

            float rotAmount = js_RotationSpeed.val * Time.deltaTime * 100;
            Quaternion _atomTargetRot;
            Transform _metaTargetTransform;

            if (rotationTargetAtom != null)
                _metaTargetTransform = rotationTargetAtom.mainController.transform;
            else
                _metaTargetTransform = positionTargetAtom.mainController.transform;

            if (_metaTargetTransform.up.y > 0)
                _atomTargetRot = Quaternion.Euler(0, _metaTargetTransform.rotation.eulerAngles.y, 0);
            else
                _atomTargetRot = Quaternion.Euler(0, _metaTargetTransform.rotation.eulerAngles.y - 180, 0);

            Vector3 _deltaTarget = _metaTargetTransform.position - targetTransform.position;
            _deltaTarget.y = 0;
            _deltaTarget.Normalize();

            Quaternion _targetRot = targetTransform.rotation;
            Quaternion _rotToTarget = Quaternion.LookRotation(_deltaTarget, Vector3.up) * Quaternion.Euler(0, js_RotationOffset.val, 0);


            switch (js_RotType.val)
            {
                case "Face Target":
                    _targetRot = Quaternion.RotateTowards(_targetRot, _rotToTarget, rotAmount);
                    break;
                case "Copy Target":
                    _targetRot = Quaternion.RotateTowards(_targetRot, _atomTargetRot * Quaternion.Euler(0, js_RotationOffset.val, 0), rotAmount);
                    break;
                case "Copy Target Idle":
                    if (isIdle)
                        _targetRot = Quaternion.RotateTowards(_targetRot, _atomTargetRot * Quaternion.Euler(0, js_RotationOffset.val, 0), rotAmount);
                    else
                        _targetRot = Quaternion.RotateTowards(_targetRot, _rotToTarget, rotAmount);
                    break;
            }

            targetTransform.rotation = _targetRot;

            float _angleToFaceTarget = Vector3.SignedAngle(targetTransform.forward, _deltaTarget.normalized, Vector3.up);

            if (js_RotType.val == "Face Target" || js_RotType.val == "Copy Target Idle")
            {
                if (Mathf.Abs(_angleToFaceTarget) > 35 + Mathf.Abs(js_RotationOffset.val))
                {
                    currentSpeed = Vector3.zero;
                    isIdle = true;
                }
            }

            movementDirection = Vector3.Dot(targetTransform.forward, _deltaTarget);
            movementDirection = Mathf.Abs(movementDirection);

            js_TargetRot.val = targetTransform.rotation.eulerAngles;
        }

        /* Getters & Setters */

        public Transform TargetTransform
        {
            get
            {
                return targetTransform;
            }
        }

        public bool UseTargetTransformRotation
        {
            get
            {
                return js_CopyAllTargetRots.val;
            }
        }

        public Atom PositionTargetAtom
        {
            get
            {
                return positionTargetAtom;
            }
        }

        public Atom RotationTargetAtom
        {
            get
            {
                if (rotationTargetAtom != null)
                    return rotationTargetAtom;
                else
                    return positionTargetAtom;
            }
        }

        public BodyController BodyController
        {
            get
            {
                return bodyController;
            }
        }

        public float GetTargetBodyHeight
        {
            get
            {
                if (positionTargetAtom != null)
                    return positionTargetAtom.mainController.transform.position.y - 0.5f;
                else
                    return 1;
            }
        }

        public Vector3 CurrentSpeed
        {
            get
            {
                return currentSpeed;
            }
        }

        public float SpeedFactor
        {
            get { return currentSpeed.magnitude / js_MaxSpeed.val; }
        }

        public float MovementDirection
        {
            get
            {
                return movementDirection;
            }
        }

        public float PresetTransitionSpeed
        {
            get { return js_PresetLoadTransitionDuration.val; }
        }

        public bool IsIdle
        {
            get
            {
                return isIdle;
            }
        }

        public bool IsStanding
        {
            get
            {
                return bodyController.IsStanding;
            }
        }

        public List<Collider> GroundColliders
        {
            get
            {               
                return groundColliders;
            }
        }

        /* Movement Functions */

        public void LockMovement()
        {
            movementLock = true;
        }

        public void UnlockMovement()
        {
            movementLock = false;
        }


        /* UI */

        protected List<UIDynamicSlider> activeSliders;
        protected List<UIDynamicToggle> activeToggles;
        protected List<UIDynamicButton> activeButtons;
        protected List<UIDynamicPopup> activePopups;
        protected List<UIDynamicTextField> activeTextFields;
        protected List<UIDynamic> activeSpacers;

        void CreateGlobalUI()
        {
            CreateMainButton("Presets", CreatePresetsUI);
            CreateMainButton("Triggers", CreateTriggersUI);

            CreateStaticText("Title", "<color=#000><size=35><b>Global</b></size></color>", false, 50);

            CreateMainButton("Movement/Rotation", CreateMovementUI);     
            
            bodyController.CreateGlobalUI();

            CreateMovementUI();
        }

        protected void CreateMovementUI()
        {
            ClearUI();

            CreateTempStaticText("Title", "<color=#000><size=35><b>Movement/Rotation</b></size></color>", 50);

            CreateTempFilterablePopup(js_PositionTargetAtom);
            CreateTempFilterablePopup(js_RotationTargetAtom);
            CreateTempFilterablePopup(js_GroundAtom);

            CreateTempToggle(js_CanMove);
            CreateTempSlider(js_MaxSpeed);
            CreateTempSlider(js_Acceleration);
            CreateTempSlider(js_StopDistance);

            CreateTempFilterablePopup(js_RotType);
            CreateTempSlider(js_RotationSpeed);
            CreateTempSlider(js_RotationOffset);
            CreateTempToggle(js_CopyAllTargetRots);
        }

        protected void CreatePresetsUI()
        {
            ClearUI();

            CreateTempStaticText("Title", "<color=#000><size=35><b>Presets</b></size></color>", 50);

            CreateTempButton("Save Preset", Callback_CreateNewFullBodyPreset);

            CreateTempButton("Load Preset", Callback_LoadFullBodyPreset);

            CreateTempSlider(js_PresetLoadTransitionDuration);
        }

        protected void CreateTriggersUI()
        {
            ClearUI();

            CreateTempStaticText("Title", "<color=#000><size=35><b>Triggers</b></size></color>", 50);

            CreateTempPopup(js_TriggerEditorChooser);

            CreateTempButton("Save Full Trigger", Callback_Trigger_SaveFullTriggerPreset);

            CreateTempButton("Load Full Trigger", Callback_Trigger_LoadFullTriggerPreset);

            CreateTempButton("Save Destination Reached", Callback_Trigger_SaveDestinationReachedPreset);

            CreateTempButton("Load Destination Reached", Callback_Trigger_LoadDestinationReachedPreset);
        }

        /* UI Helpers */

        public void CreateMainButton(string _text, UnityEngine.Events.UnityAction _callback)
        {
            ColorBlock _colors = new ColorBlock();
            _colors.normalColor = new Color(0.3f, 0.5f, 0.7f);
            _colors.pressedColor = new Color(0.2f, 0.4f, 0.6f);
            _colors.highlightedColor = new Color(0.1f, 0.1f, 0.3f);
            _colors.colorMultiplier = 1f;

            UIDynamicButton _btn = CreateButton(_text, false);
            _btn.button.onClick.AddListener(_callback);
            _btn.button.colors = _colors;

            _btn.textColor = new Color(0.9f, 0.9f, 0.9f);
            _btn.buttonText.fontSize = 35;
        }

        public void CreateTempSlider(JSONStorableFloat _data)
        {
            activeSliders.Add(CreateSlider(_data, true));
        }

        public void CreateTempToggle(JSONStorableBool _data)
        {
            activeToggles.Add(CreateToggle(_data, true));
        }

        public void CreateTempButton(string _text, UnityEngine.Events.UnityAction _callback)
        {
            UIDynamicButton _btn = CreateButton(_text, true);
            _btn.button.onClick.AddListener(_callback);

            activeButtons.Add(_btn);
        }

        public void CreateTempSpacer(float _height)
        {
            UIDynamic _spacer = CreateSpacer(true);
            _spacer.height = _height;
            activeSpacers.Add(_spacer);
        }

        public void CreateTempPopup(JSONStorableStringChooser _data)
        {
            activePopups.Add(CreatePopup(_data, true));
        }

        public void CreateTempFilterablePopup(JSONStorableStringChooser _data)
        {
            activePopups.Add(CreateFilterablePopup(_data, true));
        }

        public UIDynamicTextField CreateStaticText(string DescTitle, string DescText, bool rightSide, int fieldHeight)
        {
            JSONStorableString staticDescString = new JSONStorableString(DescTitle, DescText);
            UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
            staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
            LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
            sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
            staticDescStringField.height = fieldHeight;
            DisableScrollOnText(staticDescStringField);

            return staticDescStringField;
        }

        public UIDynamicTextField CreateTempStaticText(string DescTitle, string DescText, int fieldHeight)
        {
            activeTextFields.Add(CreateStaticText(DescTitle, DescText, true, fieldHeight));

            return activeTextFields[activeTextFields.Count - 1];
        }


        public void ClearUI()
        {
            for (int i = 0; i < activeSliders.Count; i++)
            {
                RemoveSlider(activeSliders[i]);
            }

            for (int i = 0; i < activeToggles.Count; i++)
            {
                RemoveToggle(activeToggles[i]);
            }

            for (int i = 0; i < activeButtons.Count; i++)
            {
                RemoveButton(activeButtons[i]);
            }

            for (int i = 0; i < activeSpacers.Count; i++)
            {
                RemoveSpacer(activeSpacers[i]);
            }

            for (int i = 0; i < activePopups.Count; i++)
            {
                RemovePopup(activePopups[i]);
            }
            
            for (int i = 0; i < activeTextFields.Count; i++)
            {
                RemoveTextField(activeTextFields[i]);
            }

            activeTextFields = new List<UIDynamicTextField>();
            activeSpacers = new List<UIDynamic>();
            activeButtons = new List<UIDynamicButton>();
            activeSliders = new List<UIDynamicSlider>();
            activePopups = new List<UIDynamicPopup>();
            activeToggles = new List<UIDynamicToggle>();
        }

        /* Callbacks */

        public void Callback_TargetPositionAtomChanged(string s)
        {
            List<Atom> atoms = GetSceneAtoms();

            if (s == "None")
            {
                positionTargetAtom = null;
            }
            else
            {
                foreach (Atom a in atoms)
                {
                    if (a.name == s)
                    {
                        positionTargetAtom = a;
                    }
                }
            }

        }

        public void Callback_TargetRotationAtomChanged(string s)
        {
            List<Atom> atoms = GetSceneAtoms();

            if (s == "None")
            {
                rotationTargetAtom = null;
            }
            else
            {
                foreach (Atom a in atoms)
                {
                    if (a.name == s)
                    {
                        rotationTargetAtom = a;
                    }
                }
            }

        }

        public void Callback_TargetAtomOpened()
        {
            js_PositionTargetAtom.choices = GetAtomNames();
            js_RotationTargetAtom.choices = js_PositionTargetAtom.choices;
        }

        public void Callback_GroundAtomChanged(string s)
        {
            List<Atom> atoms = GetSceneAtoms();

            if (s == "None")
            {
                groundAtom = null;
            }
            else
            {
                foreach (Atom a in atoms)
                {
                    if (a.name == s)
                    {
                        groundAtom = a;
                        groundColliders = new List<Collider>();
                        groundColliders.AddRange(groundAtom.GetComponentsInChildren<Collider>());
                    }
                }
            }
        }

        public void Callback_GroundAtomOpened()
        {
            js_GroundAtom.choices = GetAtomNames();
        }

        /* Full body presets */

        public void Callback_CreateNewFullBodyPreset()
        {
            FileManagerSecure.CreateDirectory(FBP_DIR);
            SuperController.singleton.NormalizeMediaPath(FBP_DIR + "\\");

            SuperController.singleton.GetMediaPathDialog(SaveFullBodyPreset, EXT_FULLBODY);

            // Update the browser to be a Save browser
            uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
            browser.defaultPath = FBP_DIR;
            browser.SetTextEntry(true);
            browser.fileEntryField.text = String.Format("{0}.{1}", "", EXT_FULLBODY);
            browser.ActivateFileNameField();
        }

        public void Callback_LoadFullBodyPreset()
        {
            FileManagerSecure.CreateDirectory(FBP_DIR);
            uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
            browser.defaultPath = FBP_DIR;

            SuperController.singleton.NormalizeMediaPath(FBP_DIR + "\\");
            SuperController.singleton.GetMediaPathDialog(LoadFullBodyPreset, EXT_FULLBODY);
        }

        public void SaveFullBodyPreset(string _path)
        {
            if (_path == "")
                return;

            if (!_path.ToLower().EndsWith(EXT_FULLBODY.ToLower()))
            {
                _path += "." + EXT_FULLBODY;
            }
            JSONClass _saveJSON = new JSONClass();

            try
            {
                _saveJSON = bodyController.GetAllBodyParameters();
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }

            _saveJSON["MovementParameters"] = new JSONArray();

            _saveJSON["MovementParameters"][0] = js_Acceleration.val.ToString();
            _saveJSON["MovementParameters"][1] = js_CanMove.val.ToString();
            _saveJSON["MovementParameters"][2] = js_MaxSpeed.val.ToString();
            _saveJSON["MovementParameters"][3] = js_RotationSpeed.val.ToString();
            _saveJSON["MovementParameters"][4] = js_RotType.val.ToString();
            _saveJSON["MovementParameters"][5] = js_StopDistance.val.ToString();
            _saveJSON["MovementParameters"][6] = js_PositionTargetAtom.val.ToString();
            _saveJSON["MovementParameters"][7] = js_RotationOffset.val.ToString();
            _saveJSON["MovementParameters"][8] = js_CopyAllTargetRots.val.ToString();

            SaveJSON(_saveJSON, _path);
        }

        public void LoadFullBodyPreset(string _path)
        {
            if (_path == "")
                return;

            JSONNode _jsonFile = LoadJSON(_path);

            bodyController.SetAllParametersFromSaveFile(_jsonFile);

            js_Acceleration.val = _jsonFile["MovementParameters"][0].AsFloat;
            js_CanMove.val = _jsonFile["MovementParameters"][1].AsBool;
            js_MaxSpeed.val = _jsonFile["MovementParameters"][2].AsFloat;
            js_RotationSpeed.val = _jsonFile["MovementParameters"][3].AsFloat;
            js_RotType.val = _jsonFile["MovementParameters"][4];
            js_StopDistance.val = _jsonFile["MovementParameters"][5].AsFloat;
            //js_TargetAtom.val = _jsonFile["MovementParameters"][6];
            js_RotationOffset.val = _jsonFile["MovementParameters"][7].AsFloat;
            js_CopyAllTargetRots.val = _jsonFile["MovementParameters"][8].AsBool;

            trigger_OnPresetLoadStart.Trigger();
        }


        /* Trigger Presets */
        public void Callback_Trigger_LoadFullTriggerPreset()
        {
            FileManagerSecure.CreateDirectory(TP_DIR);
            uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
            browser.defaultPath = TP_DIR;

            SuperController.singleton.NormalizeMediaPath(TP_DIR + "\\");
            SuperController.singleton.GetMediaPathDialog(LoadTriggerFullTriggerPreset, EXT_FULLTRIGGER);
        }

        public void Callback_Trigger_SaveFullTriggerPreset()
        {
            FileManagerSecure.CreateDirectory(TP_DIR);
            SuperController.singleton.NormalizeMediaPath(TP_DIR + "\\");

            SuperController.singleton.GetMediaPathDialog(SaveTriggerFullTriggerPreset, EXT_FULLTRIGGER);

            // Update the browser to be a Save browser
            uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
            browser.defaultPath = TP_DIR;
            browser.SetTextEntry(true);
            browser.fileEntryField.text = String.Format("{0}.{1}", "", EXT_FULLTRIGGER);
            browser.ActivateFileNameField();
        }

        public void SaveTriggerFullTriggerPreset(string _path)
        {
            if (_path == "")
                return;

            if (!_path.ToLower().EndsWith(EXT_FULLTRIGGER.ToLower()))
            {
                _path += "." + EXT_DEST_TRIGGER;
            }

            JSONClass _saveJSON = new JSONClass();

            _saveJSON["Triggers"] = new JSONArray();

            _saveJSON["Triggers"][0] = trigger_OnIdleStart.GetJSON().AsObject;
            _saveJSON["Triggers"][1] = trigger_OnStepIdleStart.GetJSON().AsObject;
            _saveJSON["Triggers"][2] = trigger_OnPresetLoadStart.GetJSON().AsObject;
            _saveJSON["Triggers"][3] = trigger_OnPresetLoaded.GetJSON().AsObject;
            _saveJSON["Triggers"][4] = trigger_OnMovementStart.GetJSON().AsObject;
            _saveJSON["Triggers"][5] = trigger_OnStanding.GetJSON().AsObject;
            _saveJSON["Triggers"][6] = trigger_OnKneeling.GetJSON().AsObject;
            _saveJSON["Triggers"][7] = trigger_OnDestinationReached.GetJSON().AsObject;

            SaveJSON(_saveJSON, _path);
        }

        public void LoadTriggerFullTriggerPreset(string _path)
        {
            if (_path == "")
                return;

            JSONNode _jsonFile = LoadJSON(_path);

            trigger_OnIdleStart.RestoreFromJSON(_jsonFile["Triggers"][0].AsObject);
            trigger_OnStepIdleStart.RestoreFromJSON(_jsonFile["Triggers"][1].AsObject);
            trigger_OnPresetLoadStart.RestoreFromJSON(_jsonFile["Triggers"][2].AsObject);
            trigger_OnPresetLoaded.RestoreFromJSON(_jsonFile["Triggers"][3].AsObject);
            trigger_OnMovementStart.RestoreFromJSON(_jsonFile["Triggers"][4].AsObject);
            trigger_OnStanding.RestoreFromJSON(_jsonFile["Triggers"][5].AsObject);
            trigger_OnKneeling.RestoreFromJSON(_jsonFile["Triggers"][6].AsObject);
            trigger_OnDestinationReached.RestoreFromJSON(_jsonFile["Triggers"][7].AsObject);
        }


        public void Callback_Trigger_LoadDestinationReachedPreset()
        {
            FileManagerSecure.CreateDirectory(TP_DIR);
            uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
            browser.defaultPath = TP_DIR;

            SuperController.singleton.NormalizeMediaPath(TP_DIR + "\\");
            SuperController.singleton.GetMediaPathDialog(LoadTriggerDestinationReachedPreset, EXT_DEST_TRIGGER);
        }

        public void Callback_Trigger_SaveDestinationReachedPreset()
        {
            FileManagerSecure.CreateDirectory(TP_DIR);
            SuperController.singleton.NormalizeMediaPath(TP_DIR + "\\");

            SuperController.singleton.GetMediaPathDialog(SaveTriggerDestinationReachedPreset, EXT_DEST_TRIGGER);

            // Update the browser to be a Save browser
            uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
            browser.defaultPath = TP_DIR;
            browser.SetTextEntry(true);
            browser.fileEntryField.text = String.Format("{0}.{1}", "", EXT_DEST_TRIGGER);
            browser.ActivateFileNameField();
        }

        public void SaveTriggerDestinationReachedPreset(string _path)
        {
            if (_path == "")
                return;

            if (!_path.ToLower().EndsWith(EXT_DEST_TRIGGER.ToLower()))
            {
                _path += "." + EXT_DEST_TRIGGER;
            }

            JSONClass _saveJSON = new JSONClass();

            _saveJSON["Triggers"] = new JSONArray();

            _saveJSON["Triggers"][0] = trigger_OnDestinationReached.GetJSON().AsObject;

            SaveJSON(_saveJSON, _path);
        }

        public void LoadTriggerDestinationReachedPreset(string _path)
        {
            if (_path == "")
                return;

            JSONNode _jsonFile = LoadJSON(_path);

            trigger_OnDestinationReached.RestoreFromJSON(_jsonFile["Triggers"][0].AsObject);
        }

        /* Actions Callbacks */

        public void Callback_TriggerEditorChooser(string _choice)
        {
            switch (_choice)
            {
                case "OnIdleStart":
                    trigger_OnIdleStart.OpenPanelActionStart();
                    break;

                case "OnStepIdleStart":
                    trigger_OnStepIdleStart.OpenPanelActionStart();
                    break;

                case "OnMovementStart":
                    trigger_OnMovementStart.OpenPanelActionStart();
                    break;

                case "OnStanding":
                    trigger_OnStanding.OpenPanelActionStart();
                    break;

                case "OnKneeling":
                    trigger_OnKneeling.OpenPanelActionStart();
                    break;

                case "OnBodyPresetLoadStart":
                    trigger_OnPresetLoadStart.OpenPanelActionStart();
                    break;

                case "OnBodyPresetLoaded":
                    trigger_OnPresetLoaded.OpenPanelActionStart();
                    break;

                case "OnDestinationReached":
                    trigger_OnDestinationReached.OpenPanelActionStart();
                    break;
            }

            js_TriggerEditorChooser.val = triggerList[0];
        }

        public void Event_StartIdle(bool _newState)
        {
            if (_newState)
                trigger_OnIdleStart.Trigger();
            else
                trigger_OnMovementStart.Trigger();
        }

        public void Event_DestinationReached()
        {
            trigger_OnDestinationReached.Trigger();
        }


        /* Helpers */

        public FreeControllerV3 GetControllerByName(string name)
        {
            FreeControllerV3[] ctrls = GetAllControllers();

            foreach(FreeControllerV3 ctrl in ctrls)
            {
                if (ctrl.name == name)
                    return ctrl;
            }

            return null;
        }

        public ForceReceiver GetForceReceiverByName(string name)
        {
            foreach(ForceReceiver _receiver in containingAtom.forceReceivers)
            {  
                if (_receiver.name == name)
                    return _receiver;             
            }

            return null;
        }

        public List<string> GetAtomNames()
        {
            List<Atom> atoms = GetSceneAtoms();
            List<string> atomsNames = new List<string>();

            atomsNames.Add("None");

            foreach (Atom a in atoms)
            {
                atomsNames.Add(a.name);
            }

            return atomsNames;
        }

        public static Vector3 StringToVector3(string _stringVector)
        {
            // Remove the parentheses
            if (_stringVector.StartsWith("(") && _stringVector.EndsWith(")"))
            {
                _stringVector = _stringVector.Substring(1, _stringVector.Length - 2);
            }

            // split the items
            string[] sArray = _stringVector.Split(',');

            // store as a Vector3
            Vector3 result = new Vector3(
                float.Parse(sArray[0]),
                float.Parse(sArray[1]),
                float.Parse(sArray[2]));

            return result;
        }

        private void OnAtomRename(string oldid, string newid)
        {
            foreach (EventTrigger et in eventTriggers)
            {
                et.SyncAtomNames();
            }
        }

        private static void DisableScrollOnText(UIDynamicTextField target)
        {
            ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
            if (targetSR != null)
            {
                targetSR.horizontal = false;
                targetSR.vertical = false;
            }
        }

        public static string GetPluginPath()
		{
            return "Saves\\PluginData\\JaxZoa";
		}

        /* Overrides */

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            if (includePhysical || forceStore)
            {
                needsStore = true;
                foreach (EventTrigger et in eventTriggers)
                {
                    jc[et.Name] = et.GetJSON(base.subScenePrefix);
                }

            }
            return jc;
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
            {
                foreach (EventTrigger et in eventTriggers)
                {
                    et.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
                }
            }
        }

       
    }

    public class SimpleTriggerHandler : TriggerHandler
    {
        public static bool Loaded { get; private set; }

        private static SimpleTriggerHandler myInstance;

        private RectTransform myTriggerActionsPrefab;
        private RectTransform myTriggerActionMiniPrefab;
        private RectTransform myTriggerActionDiscretePrefab;
        private RectTransform myTriggerActionTransitionPrefab;

        public static SimpleTriggerHandler Instance
        {
            get
            {
                if (myInstance == null)
                    myInstance = new SimpleTriggerHandler();
                return myInstance;
            }
        }

        public static void LoadAssets()
        {
            SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
        }

        private IEnumerator LoadAssetsInternal()
        {
            foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
                yield return x;

            Loaded = true;
        }

        private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
        {
            AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
            if (request == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
            yield return request;
            GameObject go = request.GetAsset<GameObject>();
            if (go == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
            RectTransform prefab = go.GetComponent<RectTransform>();
            if (prefab == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
            assign(prefab);
        }


        void TriggerHandler.RemoveTrigger(Trigger t)
        {
            // nothing to do
        }

        void TriggerHandler.DuplicateTrigger(Trigger t)
        {
            throw new NotImplementedException();
        }

        RectTransform TriggerHandler.CreateTriggerActionsUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionMiniUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
        }

        void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
        {
            UnityEngine.Object.Destroy(rt?.gameObject);
        }
    }

    // Wrapper for easier handling of custom triggers.
    public class EventTrigger : Trigger
    {
        public string Name
        {
            get { return name; }
            set { name = value; myNeedInit = true; }
        }

        public MVRScript Owner
        {
            get; private set;
        }

        private string name;
        private bool myNeedInit = true;

        public EventTrigger(MVRScript owner, string name)
        {
            Name = name;
            Owner = owner;
            handler = SimpleTriggerHandler.Instance;
        }

        public void OpenPanelActionStart()
        {
            if (!SimpleTriggerHandler.Loaded)
            {
                SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
                return;
            }

            triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();
            if (myNeedInit)
            {
                triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
                triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
                triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
                triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
                myNeedInit = false;
            }
        }

        public void OpenPanelTransition()
        {
            if (!SimpleTriggerHandler.Loaded)
            {
                SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
                return;
            }

            triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();

            //getRoot(triggerActionsParent,0);
            if (myNeedInit)
            {
                triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
                triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Breathing Actions";
                triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
                triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
                triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;
                triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
                myNeedInit = false;
            }
        }

        public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
        {
            if (jc.HasKey(Name))
            {
                JSONClass tc = jc[Name].AsObject;
                if (tc != null)
                    base.RestoreFromJSON(tc, subScenePrefix, isMerge);
            }
            else if (setMissingToDefault)
            {
                base.RestoreFromJSON(new JSONClass());
            }
        }

        public void Trigger()
        {
            active = true;
            active = false;
        }

        public void TriggerTransition(float val)
        {
            this.transitionInterpValue = val;
        }
    }
}